// Copyright 2016 Dietrich Epp.
// This file is part of the Moria Gamedev Repository.  This file is licensed
// under the terms of the MIT license.  For more information, see LICENSE.txt.
'use strict';
var machine = require('./machine');
var music = require('./music');
var npc = require('./npc');
var player = require('./player');
var text = require('./text');
var trigger = require('./trigger');
var util = require('./util');

var Gravity = 1500;
var Restitution = 0.1;
var ArrowMove = 32, ArrowTime = 1500;

var LevelOrder = [
	'entry1',
	'level2',
	'level3',
	'level4',
];

var Layers = [
	'wire',
	'tilebg',
	'blockbg',
	'tilefg',
	'blockfg',
	'nonplayer',
	'player',
];

function Level() {
	this.levelName = null;
	this.group = null;
	this.machine = null;
	this.npc = npc;
	this.player = null;
	this.text = null;
	this.trigger = null;
}

Level.prototype = Object.create(Phaser.State);

Level.prototype.init = function(levelName) {
	if (levelName === null || levelName === undefined) {
		levelName = util.getVar('level');
		if (levelName === undefined) {
			levelName = LevelOrder[0];
		} else if (!(levelName in Levels)) {
			util.error('No such level: ' + levelName);
		}
	}
	this.levelName = levelName;
};

Level.prototype.create = function() {
	this.group = Game.add.group();
	Game.physics.startSystem(Phaser.Physics.P2JS);
	var i, j, layer;
	if (!this.levelName) {
		util.error('No level name');
		return;
	}
	if (this.levelName.startsWith('entry')) {
		music.play('01_sunlit_cathedral');
	} else {
		music.play('02_dusty_tomb');
	}
	var level = Levels[this.levelName];
	if (!level) {
		util.error('No such level: ' + this.levelName);
		return;
	}
	var layers = {};
	for (i = 0; i < Layers.length; i++) {
		layers[Layers[i]] = [];
	}
	var map = Game.make.tilemap(null, 32, 32);
	map.addTilesetImage('tile');
	var lw = level.width, lh = level.height;
	var bg = map.createBlankLayer('bg', lw, lh, 32, 32, this.group);
	layers.tilebg.push(bg);
	var fg = map.createBlankLayer('fg', lw, lh, 32, 32, this.group);
	layers.tilefg.push(fg);
	var data = level.tiles.split(',');
	for (i = 0; i < data.length; i++) {
		var tstr = data[i];
		if (tstr === '') {
			continue;
		}
		var tile = parseInt(tstr);
		layer = fg;
		if ((tile >= 1 && tile <= 7)) {
			layer = bg;
		}
		map.putTile(tile, i % lw, (i / lw) | 0, layer);
	}
	fg.resizeWorld();
	var coll = [];
	for (i = 8; i < 16; i++) { coll.push(i); }
	for (i = 18; i < 24; i++) { coll.push(i); }
	for (i = 24; i < 27; i++) { coll.push(i); }
	for (i = 32; i < 35; i++) { coll.push(i); }
	map.setCollision(coll, true, fg);
	Game.physics.p2.convertTilemap(map, fg);
	Game.physics.p2.restitution = Restitution;
	Game.physics.p2.gravity.y = Gravity;
	var objects = level.objects;
	this.player = null;
	this.machine = new machine.Machine(level);
	this.npc = new npc.NPC(level);
	this.text = new text.Text(level);
	this.trigger = new trigger.Trigger(level);
	for (i = 0; i < objects.length; i++) {
		var od = objects[i];
		switch (od.type) {
		case 'player':
			if (this.player !== null) {
				util.warn('Multiple players');
				break;
			}
			this.player = new player.Player(od, this.group, layers);
			break;
		case 'door':
		case 'plate':
		case 'platform':
		case 'autoplat':
		case 'wire':
			this.machine.add(od, this.group, layers);
			break;
		case 'npc':
			this.npc.add(od, this.group, layers);
			break;
		case 'trigger':
		case 'teleport':
		case 'teleport_target':
			this.trigger.add(od, this.group, layers);
			break;
		case 'arrow':
			this.addArrow(od, this.group, layers);
			break;
		default:
			util.warn('Unknown object type: ' + od.type);
			break;
		}
	}
	for (i = 0; i < Layers.length; i++) {
		layer = layers[Layers[i]];
		for (j = 0; j < layer.length; j++) {
			this.group.bringToTop(layer[j]);
		}
	}
	if (this.player === null) {
		util.error('No player');
		return;
	}
	this.player.create();
	this.trigger.create();
	this.npc.create();
	this.machine.create();
	this.text.create();
};

Level.prototype.addArrow = function(od, group, layers) {
	var x = od.x * 32 + 16, y = od.y * 32 - 16;
	var sp = Game.add.sprite(x - ArrowMove / 2, y, 'tile', 'i.arrow');
	sp.alpha = 0;
	sp.anchor.set(0.5, 0.5);
	layers.nonplayer.push(sp);
	var tween = Game.add.tween(sp)
			.to({x: x, alpha: 1}, ArrowTime / 2)
			.to({x: x + ArrowMove / 2, alpha: 0}, ArrowTime / 2)
			.loop(true)
			.start();
};

Level.prototype.update = function() {
	this.player.update();
	this.trigger.update(this);
	this.npc.update();
	this.machine.update();
	this.text.update();
};

Level.prototype.exit = function() {
	var i = LevelOrder.indexOf(this.levelName);
	if (i >= 0) {
		i++;
		if (i >= LevelOrder.length) {
			Game.state.start('End', true, false);
			return;
		}
	} else {
		i = 0;
	}
	Game.state.start('Level', true, false, LevelOrder[i]);
};

module.exports = {
	Level: Level,
};
