// Copyright 2016 Dietrich Epp.
// This file is part of the Moria Gamedev Repository.  This file is licensed
// under the terms of the MIT license.  For more information, see LICENSE.txt.
'use strict';
var util = require('./util');

var PlatePressTime = 150;
var AutoplatPressTime = 500;
var PlatformPause = 250;

function trigger(devices, active) {
	for (var i = 0; i < devices.length; i++) {
		devices[i].trigger(active);
	}
}

////////////////////////////////////////////////////////////////////////////////

function Lights(color) {
	this.sprites = [];
	this.onFrames = [];
	this.offFrames = [];
	this.color = color;
}

Lights.prototype.trigger = function(active) {
	var sprites = this.sprites, frames = active ? this.onFrames : this.offFrames;
	for (var i = 0; i < sprites.length; i++) {
		sprites[i].frameName = frames[i];
	}
};

Lights.prototype.add = function(sprites) {
	for (var i = 0; i < sprites.length; i++) {
		var s = sprites[i];
		this.sprites.push(s);
		var f = s.frameName, fon, foff;
		if (f.startsWith('b')) {
			fon = 'b' + this.color + 'b';
			foff = 'b' + this.color + 'a';
		} else if (f.startsWith('w')) {
			fon = 'w' + this.color + f[2];
			foff = 'w0' + f[2];
		}
		s.frameName = foff;
		this.onFrames.push(fon);
		this.offFrames.push(foff);
	}
};

////////////////////////////////////////////////////////////////////////////////

function Signal(name) {
	this.index = name;
	this.devices = [];
	this.inputCount = 0;
	this.active = false;
	this.lights = null;
	this.color = name;
}

Signal.prototype.update = function() {
	var active = !!this.inputCount;
	if (active !== this.active) {
		this.active = active;
		trigger(this.devices, active);
	}
};

Signal.prototype.trigger = function(active) {
	if (active) {
		this.inputCount++;
	} else {
		this.inputCount--;
	}
};

Signal.prototype.getLights = function() {
	if (!this.lights) {
		this.lights = new Lights(this.color);
		this.devices.push(this.lights);
	}
	return this.lights;
};

////////////////////////////////////////////////////////////////////////////////

function Plate(offDelay, invert) {
	this.pressTime = -1;
	this.active = false;
	this.devices = [];
	this.offDelay = offDelay;
	this.invert = !!invert;
}

Plate.prototype.onPress = function() {
	var now = Game.time.now;
	this.pressTime = now + this.offDelay;
};

Plate.prototype.update = function() {
	var now = Game.time.now;
	var active = now < this.pressTime;
	if (this.invert) {
		active = !active;
	}
	if (active !== this.active) {
		Game.sound.play('sfx/click' + (Math.random() > 0.5 ? '1': '2'));
		this.active = active;
		trigger(this.devices, active);
	}
};

////////////////////////////////////////////////////////////////////////////////

// Args: group, collider, dir, travel, speed
function Mover(group, collider, od) {
	this.group = group;
	this.collider = collider;
	var dir = od.dir;
	if (dir === undefined || dir === null) {
		dir = 'up';
	}
	var travel = od.travel;
	if (travel === undefined || travel === null) {
		var lip = od.lip;
		if (lip === undefined || lip === null) {
			lip = 16;
		}
		switch (dir) {
		case 'up':
		case 'down':
			travel = collider.height - lip;
			break;
		case 'left':
		case 'right':
			travel = collider.width - lip;
			break;
		}
	}
	this.x0 = this.group.x;
	this.y0 = this.group.y;
	this.x1 = this.x0;
	this.y1 = this.y0;
	switch (dir) {
	case 'up':
		this.y1 -= travel;
		break;
	case 'down':
		this.y1 += travel;
		break;
	case 'left':
		this.x1 -= travel;
		break;
	case 'right':
		this.x1 += travel;
		break;
	}
	this.dir = dir;
}

Mover.prototype.getPhase = function() {
	switch (this.dir) {
	case 'up':
	case 'down':
		return (this.collider.y - this.y0) / (this.y1 - this.y0);
	case 'left':
	case 'right':
		return (this.collider.x - this.x0) / (this.x1 - this.x0);
	}
};

Mover.prototype.update = function(velocity) {
	var v = this.collider.body.velocity;
	switch (this.dir) {
	case 'up':
		v.y = -velocity;
		break;
	case 'down':
		v.y = velocity;
		break;
	case 'left':
		v.x = -velocity;
		break;
	case 'right':
		v.x = velocity;
		break;
	}
	this.group.x = this.collider.x;
	this.group.y = this.collider.y;
};

////////////////////////////////////////////////////////////////////////////////

function Door(group, collider, od) {
	this.mover = new Mover(group, collider, od);
	this.active = false;
	this.speed = 100;
	if ('speed' in od) {
		this.speed *= od.speed;
	}
	this.delay = 0;
	if ('delay' in od) {
		this.delay = od.delay;
	}
	this.delayTime = -1;
	this.sound = null;
}

Door.prototype.update = function() {
	var phase = this.mover.getPhase();
	var vel = 0;
	if (this.active && Game.time.now > this.delayTime) {
		if (phase < 0.9) {
			vel = this.speed;
		} else if (phase < 0.99) {
			vel = this.speed * 0.5;
		}
	} else {
		if (phase > 0.1) {
			vel = -this.speed;
		} else if (phase > 0.01) {
			vel = -this.speed * 0.5;
		}
	}
	if (Math.abs(vel) > 0.01) {
		if (!this.sound) {
			this.sound = Game.sound.play('sfx/scrape');
		}
	} else {
		if (this.sound) {
			this.sound.stop();
		}
		this.sound = null;
	}
	this.mover.update(vel);
};

Door.prototype.trigger = function(active) {
	this.active = active;
	if (active && this.delay > 0) {
		this.delayTime = Game.time.now + this.delay;
	}
};

////////////////////////////////////////////////////////////////////////////////

function Platform(group, collider, od) {
	this.mover = new Mover(group, collider, od);
	this.active = false;
	this.speed = 150;
	if ('speed' in od) {
		this.speed *= od.speed;
	}
	this.pause = PlatformPause;
	if ('pause' in od) {
		this.pause = 1000 * od.pause;
	}
	this.delay = 0;
	if ('delay' in od) {
		this.delay = od.delay;
	}
	this.reverse = false;
	this.pauseTime = -1;
}

Platform.prototype.update = function() {
	var phase = this.mover.getPhase();
	var vel = 0;
	var now = Game.time.now;
	if (this.active && now > this.pauseTime) {
		if (!this.reverse) {
			if (phase < 0.9) {
				vel = this.speed;
			} else if (phase < 1.0) {
				vel = this.speed * 0.5;
			} else {
				this.reverse = true;
				this.pauseTime = now + this.pause;
			}
		} else {
			if (phase > 0.1) {
				vel = -this.speed;
			} else if (phase > 0.0) {
				vel = -this.speed * 0.5;
			} else {
				this.reverse = false;
				this.pauseTime = now + this.pause;
			}
		}
	}
	this.mover.update(vel);
};

Platform.prototype.trigger = function(active) {
	this.active = active;
	if (active && this.delay > 0) {
		this.pauseTime = Game.time.now + this.delay;
	}
};

////////////////////////////////////////////////////////////////////////////////

function Machine(levelData) {
	this.signals = {};
	this.parts = [];
	if (levelData.activeSignals) {
		for (var a = levelData.activeSignals, i = 0; i < a.length; i++) {
			this.getSignal(a[i]).trigger(true);
		}
	}
}

Machine.prototype.create = function() {};

Machine.prototype.update = function() {
	var i, a;
	for (a = this.parts, i = 0; i < a.length; i++) {
		a[i].update();
	}
	a = this.signals;
	for (i in a) {
		a[i].update();
	}
};

Machine.prototype.getSignal = function(name) {
	var signal = this.signals[name];
	if (!signal) {
		signal = new Signal(name);
		this.signals[name] = signal;
	}
	return signal;
};

Machine.prototype.add = function(od, group, layers) {
	switch (od.type) {
	case 'door':
	case 'plate':
	case 'platform':
	case 'autoplat':
		this.addBlock(od, group, layers);
		break;
	case 'wire':
		this.addWire(od, group, layers);
		break;
	default:
		util.warn('Unknown object type: ' + od.type);
		break;
	}
};

Machine.prototype.addBlock = function(od, group, layers) {
	var w = od.w, h = od.h;
	var x = od.x * 32 + w * 16, y = od.y * 32 + h * 16;
	var block = Game.add.group(group);
	block.x = x;
	block.y = y;
	var collider = Game.add.sprite(x, y, 'tile', 'blank', group);
	collider.anchor.set(0.5, 0.5);
	collider.width = w * 32;
	collider.height = h * 32;
	Game.physics.p2.enable(collider);
	collider.body.kinematic = true;
	var sprites = [];
	for (var bx = 0; bx < w; bx++) {
		for (var by = 0; by < h; by++) {
			var tile = Game.add.sprite(
				bx * 32 - w * 16 + 16,
				by * 32 - h * 16 + 16,
				'tile',
				'b1b',
				block);
			tile.anchor.set(0.5, 0.5);
			sprites.push(tile);
		}
	}
	var sig = this.getSignal(od.signal);
	var obj, obj2, isInput = false;
	switch (od.type) {
	case 'door':
		layers.blockbg.push(block);
		obj = new Door(block, collider, od);
		this.parts.push(obj);
		break;
	case 'platform':
		layers.blockfg.push(block);
		obj = new Platform(block, collider, od);
		this.parts.push(obj);
		break;
	case 'plate':
		isInput = true;
		layers.blockfg.push(block);
		obj = new Plate(od.latch ? Infinity : PlatePressTime, od.invert);
		obj.devices.push(sig);
		collider.data.onPress = obj.onPress.bind(obj);
		this.parts.push(obj);
		break;
	case 'autoplat':
		isInput = true;
		layers.blockfg.push(block);
		obj = new Plate(AutoplatPressTime);
		collider.data.onPress = obj.onPress.bind(obj);
		obj2 = new Door(block, collider, od);
		this.parts.push(obj, obj2);
		obj.devices.push(obj2);
		break;
	default:
		util.warn('Unknown block type: ' + od.type);
		return;
	}
	var lights;
	if (isInput) {
		lights = new Lights(sig.color);
		obj.devices.push(lights);
	} else {
		lights = sig.getLights();
		sig.devices.push(obj);
	}
	lights.add(sprites);
};

Machine.prototype.addWire = function(od, group, layers) {
	var wire = Game.add.group(group);
	layers.wire.push(wire);
	var segs = od.segments;
	var sprites = [];
	for (var i = 0; i < segs.length; i++) {
		var s = segs[i], x1 = s[0], y1 = s[1], x2 = s[2], y2 = s[3];
		var c0, c1, c, tile;
		if (x1 == x2) {
			c0 = Math.min(y1, y2);
			c1 = Math.max(y1, y2);
			for (c = c0; c < c1; c++) {
				tile = Game.add.sprite(
					x1 * 32,
					c * 32 + 16,
					'tile',
					'w1v',
					wire);
				tile.anchor.set(0.5, 0.5);
				sprites.push(tile);
			}
		} else {
			c0 = Math.min(x1, x2);
			c1 = Math.max(x1, x2);
			for (c = c0; c < c1; c++) {
				tile = Game.add.sprite(
					c * 32 + 16,
					y1 * 32,
					'tile',
					'w1h',
					wire);
				tile.anchor.set(0.5, 0.5);
				sprites.push(tile);
			}
		}
	}
	this.getSignal(od.signal).getLights().add(sprites);
};

module.exports = {
	Machine: Machine,
};
